"""Plugins for testing purposes.

Source them like this from within a test function:

api.deregister_all_paths()
api.register_plugin_path(os.path.dirname(__file__))

This ensures that the plugins are actually loaded through `plugin.discover`.
"""
from pyblish import api


class FailingExplicitPlugin(api.InstancePlugin):
    """Raise an exception."""

    def process(self, instance):
        raise Exception("A test exception")


class FailingImplicitPlugin(api.Validator):
    """Raise an exception."""

    def process(self, instance):
        raise Exception("A test exception")




# Local library
from .. import lib

import pyblish.plugin
import pyblish.logic

from nose.tools import (
    assert_equals,
    assert_true,
    assert_false,
    with_setup
)


@with_setup(lib.setup_empty, lib.teardown)
def test_process_callables():
    """logic.process can take either data or callables"""
    count = {"#": 0}

    class SelectInstance(pyblish.api.Selector):
        def process(self, context):
            instance = context.create_instance("MyInstance")
            instance.set_data("family", "myFamily")
            count["#"] += 1

    class ValidateInstance(pyblish.api.Validator):
        def process(self, instance):
            count["#"] += 10
            assert False, "I was programmed to fail"

    class ExtractInstance(pyblish.api.Extractor):
        def process(self, instance):
            count["#"] += 100

    pyblish.api.register_plugin(SelectInstance)
    pyblish.api.register_plugin(ValidateInstance)
    pyblish.api.register_plugin(ExtractInstance)

    for result in pyblish.logic.process(
            func=pyblish.plugin.process,
            plugins=pyblish.plugin.discover(),
            context=pyblish.plugin.Context()):

        if isinstance(result, pyblish.logic.TestFailed):
            break

        if isinstance(result, Exception):
            assert False  # This would be a bug

    assert_equals(count["#"], 11)

    context = pyblish.plugin.Context()

    def _context():
        return context

    count["#"] = 0

    for result in pyblish.logic.process(
            func=pyblish.plugin.process,
            plugins=pyblish.plugin.discover,  # <- Callable
            context=_context):  # <- Callable

        if isinstance(result, pyblish.logic.TestFailed):
            break

        if isinstance(result, Exception):
            assert False  # This would be a bug

    assert_equals(count["#"], 11)


@with_setup(lib.setup_empty, lib.teardown)
def test_repair():
    """Repairing with DI works well"""

    _data = {}

    class SelectInstance(pyblish.api.Selector):
        def process(self, context):
            instance = context.create_instance("MyInstance")
            instance.set_data("family", "MyFamily")

    class ValidateInstance(pyblish.api.Validator):
        def process(self, instance):
            _data["broken"] = True
            assert False, "Broken"

        def repair(self, instance):
            _data["broken"] = False

    context = pyblish.api.Context()

    results = list()
    for result in pyblish.logic.process(
            func=pyblish.plugin.process,
            plugins=[SelectInstance, ValidateInstance],
            context=context):

        if isinstance(result, pyblish.logic.TestFailed):
            assert str(result) == "Broken"

        results.append(result)

    assert_true(_data["broken"])

    repair = list()
    for result in results:
        if result["error"]:
            repair.append(result["plugin"])

    for result in pyblish.logic.process(
            func=pyblish.plugin.repair,
            plugins=repair,
            context=context):
        print(result)

    assert_false(_data["broken"])


@with_setup(lib.setup_empty, lib.teardown)
def test_context_once():
    """Context is only processed once, with DI"""

    count = {"#": 0}

    class SelectMany(pyblish.api.Selector):
        def process(self, context):
            for name in ("A", "B", "C"):
                instance = context.create_instance(name)
                instance.set_data("family", "myFamily")

    class ValidateContext(pyblish.api.Validator):
        families = ["*", "myFamily"]

        def process(self, context):
            count["#"] += 1

    pyblish.util.publish(plugins=[SelectMany, ValidateContext])
    assert_equals(count["#"], 1)


@with_setup(lib.setup_empty, lib.teardown)
def test_incompatible_context():
    """Context is processed regardless of families"""

    count = {"#": 0}

    class SelectMany(pyblish.api.Selector):
        def process(self, context):
            for name in ("A", "B", "C"):
                instance = context.create_instance(name)
                instance.set_data("family", "myFamily")

    class ValidateContext(pyblish.api.Validator):
        def process(self, context):
            count["#"] += 1

    pyblish.util.publish(plugins=[SelectMany, ValidateContext])
    assert_equals(count["#"], 1)

    count["#"] = 0

    # Even with a family, if an instance is not requested
    # it still processes.
    class ValidateContext(pyblish.api.Validator):
        families = ["NOT_EXIST"]

        def process(self, context):
            count["#"] += 1

    pyblish.util.publish(plugins=[SelectMany, ValidateContext])
    assert_equals(count["#"], 1)

    count["#"] = 0

    # However, when an instance is requested,
    # do nothing.
    class ValidateContext(pyblish.api.Validator):
        families = ["NOT_EXIST"]

        def process(self, instance):
            count["#"] += 1

    pyblish.util.publish(plugins=[SelectMany, ValidateContext])
    assert_equals(count["#"], 0)


@with_setup(lib.setup_empty, lib.teardown)
def test_custom_test():
    """Registering a custom test works fine"""

    count = {"#": 0}

    def custom_test(**vars):
        print("I accept anything")
        return

    class MyValidator(pyblish.api.Validator):
        def process(self, context):
            assert False, "I won't stop the extractor"

    class MyExtractor(pyblish.api.Extractor):
        def process(self, context):
            print("I came, I saw, I extracted..")
            count["#"] += 1

    pyblish.api.register_test(custom_test)
    pyblish.util.publish(plugins=[MyValidator, MyExtractor])
    assert_equals(count["#"], 1)


def test_logic_process():
    """logic.process works fine"""

    context = pyblish.api.Context()
    provider = pyblish.plugin.Provider()
    provider.inject("context", context)

    def my_process(plugin, context, instance=None):
        result = {
            "success": False,
            "plugin": plugin,
            "instance": None,
            "error": None,
            "records": list(),
            "duration": None
        }

        plugin = plugin()
        provider = pyblish.plugin.Provider()
        provider.inject("context", context)
        provider.invoke(plugin.process)
        return result

    class SelectInstance(pyblish.api.Selector):

        def process(self, context):
            context.create_instance("MyInstance")

    context = pyblish.util.publish(plugins=[SelectInstance])
    assert not isinstance(context.data["results"][0]["error"],
                          pyblish.logic.TestFailed)
    assert_equals(len(context), 1)


@with_setup(lib.setup_empty, lib.teardown)
def test_active():
    """An inactive plug-in won't be processed by default logic"""

    count = {"#": 0}

    class SelectInstances4321(pyblish.api.Selector):
        def process(self, context):
            for name in ("A", "B"):
                instance = context.create_instance(name)
                instance.set_data("family", "myFamily")

            count["#"] += 1

    class ValidateNotActive(pyblish.api.Validator):
        """Doesn't run, due to being inactive"""
        active = False

        def process(self, instance):
            count["#"] += 10
            assert False

    class ValidateActive(pyblish.api.Validator):
        """Runs twice, for both instances"""
        active = True

        def process(self, instance):
            count["#"] += 100

    for plugin in (SelectInstances4321, ValidateNotActive, ValidateActive):
        pyblish.api.register_plugin(plugin)

    pyblish.util.publish()
    assert_equals(count["#"], 201)


@with_setup(lib.setup_empty, lib.teardown)
def test_failing_selector():
    """Failing Selector should not abort publishing"""

    count = {"#": 0}

    class MySelector(pyblish.api.Selector):
        def process(self, context):
            assert False, "I shouldn't stop Extraction"

    class MyExtractor(pyblish.api.Extractor):
        def process(self):
            count["#"] += 1

    pyblish.api.register_plugin(MySelector)
    pyblish.api.register_plugin(MyExtractor)

    pyblish.util.publish()
    assert_equals(count["#"], 1)


@with_setup(lib.setup_empty, lib.teardown)
def test_decrementing_order():
    """Decrementing order works fine"""

    count = {"#": 0}

    class MyDecrementingSelector(pyblish.api.Selector):
        order = pyblish.api.Selector.order - 0.3

        def process(self):
            count["#"] += 0.1

    class MySelector(pyblish.api.Selector):
        def process(self, context):
            count["#"] += 1
            assert False, "I shouldn't stop Extraction"

    class MyValidator(pyblish.api.Validator):
        order = pyblish.api.Validator.order - 0.4

        def process(self):
            count["#"] += 10
            assert False, "I will run first, and stop things"

    class MyValidator2(pyblish.api.Validator):
        order = pyblish.api.Validator.order + 0.4

        def process(self):
            count["#"] += 100
            assert False, "I will stop things"

    class MyExtractor(pyblish.api.Extractor):
        order = pyblish.api.Extractor.order - 0.49

        def process(self):
            count["#"] += 1000
            assert False, "I will not run"

    class MyExtractor2(pyblish.api.Extractor):
        order = pyblish.api.Extractor.order - 0.2

        def process(self):
            count["#"] += 10000
            assert False, "I will not run"

    plugins = [
        MyDecrementingSelector,
        MySelector,
        MyValidator,
        MyValidator2,
        MyExtractor
    ]

    pyblish.util.publish(plugins=plugins)
    assert_equals(count["#"], 111.1)


def test_test():
    """The test halts an invalid publish"""

    count = {"#": 0}

    class Collector(pyblish.api.Collector):
        def process(self):
            count["#"] += 1

    class Validator(pyblish.api.Validator):
        def process(self):
            count["#"] += 10
            assert False, "I will run first, and stop things"

    class Extractor(pyblish.api.Extractor):
        def process(self):
            count["#"] += 1000
            assert False, "I will not run"

    pyblish.util.publish(plugins=[Collector, Validator, Extractor])
    assert count["#"] == 11


def test_extract_traceback():
    e = None

    try:
        1 / 0
    except Exception as e:
        assert not hasattr(e, "traceback")
        pyblish.logic._extract_traceback(e)
        assert hasattr(e, "traceback")


def test_plugins_by_families():
    """The right plug-ins are returned from plugins_by_families"""

    class ClassA(pyblish.api.Collector):
        families = ["a"]

    class ClassB(pyblish.api.Collector):
        families = ["b"]

    class ClassC(pyblish.api.Collector):
        families = ["c"]

    assert pyblish.logic.plugins_by_families(
        [ClassA, ClassB, ClassC], ["a"]) == [ClassA]



"""Convenience functions for general publishing"""

from __future__ import absolute_import

# Standard library
import logging
import warnings

# Local library
from . import api, logic, plugin, lib

log = logging.getLogger("pyblish.util")

__all__ = [
    "publish",
    "collect",
    "validate",
    "extract",
    "integrate",

    # Iterator counterparts
    "publish_iter",
    "collect_iter",
    "validate_iter",
    "extract_iter",
    "integrate_iter",
]


def publish(context=None, plugins=None, targets=None):
    """Publish everything

    This function will process all available plugins of the
    currently running host, publishing anything picked up
    during collection.

    Arguments:
        context (Context, optional): Context, defaults to
            creating a new context
        plugins (list, optional): Plug-ins to include,
            defaults to results of discover()
        targets (list, optional): Targets to include for publish session.

    Returns:
        Context: The context processed by the plugins.

    Usage:
        >> context = plugin.Context()
        >> publish(context)     # Pass..
        >> context = publish()  # ..or receive a new

    """

    context = context if context is not None else api.Context()

    for _ in publish_iter(context, plugins, targets):
        pass

    return context


def publish_iter(context=None, plugins=None, targets=None):
    """Publish iterator

    This function will process all available plugins of the
    currently running host, publishing anything picked up
    during collection.

    Arguments:
        context (Context, optional): Context, defaults to
            creating a new context
        plugins (list, optional): Plug-ins to include,
            defaults to results of discover()
        targets (list, optional): Targets to include for publish session.

    Yields:
        tuple of dict and Context: A tuple is returned with a dictionary and
            the Context object. The dictionary contains all the result
            information of a plugin process, and the Context is the Context
            after the plugin has been processed.

    Usage:
        >> context = plugin.Context()
        >> for result in util.publish_iter(context):
               print result
        >> for result in util.publish_iter():
               print result

    """
    for result in _convenience_iter(context, plugins, targets):
        yield result

    api.emit("published", context=context)


def _convenience_iter(context=None, plugins=None, targets=None, order=None):
    # Must check against None, as objects be emptys
    context = api.Context() if context is None else context
    plugins = api.discover() if plugins is None else plugins

    if order is not None:
        plugins = list(
            Plugin for Plugin in plugins
            if lib.inrange(Plugin.order, order)
        )

    # Do not consider inactive plug-ins
    plugins = list(p for p in plugins if p.active)
    collectors = list(p for p in plugins if lib.inrange(
        number=p.order,
        base=api.CollectorOrder)
    )

    # Compute an approximation of all future tasks
    # NOTE: It's an approximation, because tasks are
    # dynamically determined at run-time by contents of
    # the context and families of contained instances;
    # each of which may differ between task.
    task_count = len(list(logic.Iterator(plugins, context, targets=targets)))

    # First pass, collection
    tasks_processed_count = 1
    for Plugin, instance in logic.Iterator(collectors,
                                           context,
                                           targets=targets):
        result = plugin.process(Plugin, context, instance)

        # Inject additional member for results here.
        result["progress"] = float(tasks_processed_count) / task_count

        tasks_processed_count += 1
        yield result

    # Exclude collectors from further processing
    plugins = list(p for p in plugins if p not in collectors)

    # Exclude plug-ins that do not have at
    # least one compatible instance.
    for Plugin in list(plugins):
        if Plugin.__instanceEnabled__:
            if not logic.instances_by_plugin(context, Plugin):
                plugins.remove(Plugin)

    # Mutable state, used in Iterator
    state = {
        "nextOrder": None,
        "ordersWithError": set()
    }

    # Second pass, the remainder
    for Plugin, instance in logic.Iterator(plugins,
                                           context,
                                           state,
                                           targets=targets):
        try:
            result = plugin.process(Plugin, context, instance)
            result["progress"] = (
                float(tasks_processed_count) / task_count
            )

            tasks_processed_count += 1
        except StopIteration:  # End of items
            raise

        except Exception:  # This is unexpected, most likely a bug
            log.error("An expected exception occurred.\n")
            raise

        else:
            # Make note of the order at which the
            # potential error error occured.
            if result["error"]:
                state["ordersWithError"].add(Plugin.order)

        if isinstance(result, Exception):
            log.error("An unexpected error happened: %s" % result)
            break

        error = result["error"]
        if error is not None:
            print(error)

        yield result


def collect(context=None, plugins=None, targets=None):
    """Convenience function for collection-only

     _________    . . . . .  .   . . . . . .   . . . . . . .
    |         |   .          .   .         .   .           .
    | Collect |-->. Validate .-->. Extract .-->. Integrate .
    |_________|   . . . . .  .   . . . . . .   . . . . . . .

    """

    context = context if context is not None else api.Context()
    for result in collect_iter(context, plugins, targets):
        pass

    return context


def validate(context=None, plugins=None, targets=None):
    """Convenience function for validation-only

    . . . . . .    __________    . . . . . .   . . . . . . .
    .         .   |          |   .         .   .           .
    . Collect .-->| Validate |-->. Extract .-->. Integrate .
    . . . . . .   |__________|   . . . . . .   . . . . . . .

    """

    context = context if context is not None else api.Context()
    for result in validate_iter(context, plugins, targets):
        pass

    return context


def extract(context=None, plugins=None, targets=None):
    """Convenience function for extraction-only

    . . . . . .   . . . . .  .    _________    . . . . . . .
    .         .   .          .   |         |   .           .
    . Collect .-->. Validate .-->| Extract |-->. Integrate .
    . . . . . .   . . . . .  .   |_________|   . . . . . . .

    """

    context = context if context is not None else api.Context()
    for result in extract_iter(context, plugins, targets):
        pass

    return context


def integrate(context=None, plugins=None, targets=None):
    """Convenience function for integration-only

    . . . . . .   . . . . .  .   . . . . . .    ___________
    .         .   .          .   .         .   |           |
    . Collect .-->. Validate .-->. Extract .-->| Integrate |
    . . . . . .   . . . . .  .   . . . . . .   |___________|

    """

    context = context if context is not None else api.Context()
    for result in integrate_iter(context, plugins, targets):
        pass

    return context


def collect_iter(context=None, plugins=None, targets=None):
    for result in _convenience_iter(context, plugins, targets,
                                    order=api.CollectorOrder):
        yield result

    api.emit("collected", context=context)


def validate_iter(context=None, plugins=None, targets=None):
    for result in _convenience_iter(context, plugins, targets,
                                    order=api.ValidatorOrder):
        yield result

    api.emit("validated", context=context)


def extract_iter(context=None, plugins=None, targets=None):
    for result in _convenience_iter(context, plugins, targets,
                                    order=api.ExtractorOrder):
        yield result

    api.emit("extracted", context=context)


def integrate_iter(context=None, plugins=None, targets=None):
    for result in _convenience_iter(context, plugins, targets,
                                    order=api.IntegratorOrder):
        yield result

    api.emit("integrated", context=context)


def _convenience(context=None, plugins=None, targets=None, order=None):
    context = context if context is not None else api.Context()

    for result in _convenience_iter(context, plugins, targets, order):
        pass

    return context


# Backwards compatibility
select = collect
conform = integrate
run = publish  # Alias


def publish_all(context=None, plugins=None):
    warnings.warn("pyblish.util.publish_all has been "
                  "deprecated; use publish()")
    return publish(context, plugins)


def validate_all(context=None, plugins=None):
    warnings.warn("pyblish.util.validate_all has been "
                  "deprecated; use collect() followed by validate()")
    context = collect(context, plugins)
    return validate(context, plugins)



import pyblish.api
import pyblish.util
from nose.tools import (
    with_setup,
)
from . import lib


@with_setup(lib.setup_empty)
def test_published_event():
    """published is emitted upon finished publish"""

    count = {"#": 0}

    def on_published(context):
        assert isinstance(context, pyblish.api.Context)
        count["#"] += 1

    pyblish.api.register_callback("published", on_published)
    pyblish.util.publish()

    assert count["#"] == 1, count


@with_setup(lib.setup_empty)
def test_validated_event():
    """validated is emitted upon finished validation"""

    count = {"#": 0}

    def on_validated(context):
        assert isinstance(context, pyblish.api.Context)
        count["#"] += 1

    pyblish.api.register_callback("validated", on_validated)
    pyblish.util.validate()

    assert count["#"] == 1, count

@with_setup(lib.setup_empty)
def test_plugin_processed_event():
    """pluginProcessed is emitted upon a plugin being processed, regardless of its success"""

    class MyContextCollector(pyblish.api.ContextPlugin):
        order = pyblish.api.CollectorOrder

        def process(self, context):
            context.create_instance("A")

    class CheckInstancePass(pyblish.api.InstancePlugin):
        order = pyblish.api.ValidatorOrder

        def process(self, instance):
            pass

    class CheckInstanceFail(pyblish.api.InstancePlugin):
        order = pyblish.api.ValidatorOrder

        def process(self, instance):
            raise Exception("Test Fail")

    pyblish.api.register_plugin(MyContextCollector)
    pyblish.api.register_plugin(CheckInstancePass)
    pyblish.api.register_plugin(CheckInstanceFail)


    count = {"#": 0}

    def on_processed(result):
        assert isinstance(result, dict)
        count["#"] += 1

    pyblish.api.register_callback("pluginProcessed", on_processed)
    pyblish.util.publish()

    assert count["#"] == 3, count

@with_setup(lib.setup_empty)
def test_plugin_failed_event():
    """pluginFailed is emitted upon a plugin failing for any reason"""

    class MyContextCollector(pyblish.api.ContextPlugin):
        order = pyblish.api.CollectorOrder
        def process(self, context):
            context.create_instance("A")

    class CheckInstancePass(pyblish.api.InstancePlugin):
        order = pyblish.api.ValidatorOrder
        def process(self, instance):
            pass

    class CheckInstanceFail(pyblish.api.InstancePlugin):
        order = pyblish.api.ValidatorOrder
        def process(self, instance):
            raise Exception("Test Fail")

    pyblish.api.register_plugin(MyContextCollector)
    pyblish.api.register_plugin(CheckInstancePass)
    pyblish.api.register_plugin(CheckInstanceFail)

    count = {"#": 0}

    def on_failed(plugin, context, instance, error):
        assert issubclass(plugin, pyblish.api.InstancePlugin) #plugin == CheckInstanceFail
        assert isinstance(context, pyblish.api.Context)
        assert isinstance(instance, pyblish.api.Instance)
        assert isinstance(error, Exception)

        count["#"] += 1

    pyblish.api.register_callback("pluginFailed", on_failed)
    pyblish.util.publish()

    assert count["#"] == 1, count


"""Pyblish command-line interface

Usage:
    $ pyblish --help

Attributes:
    SCREEN_WIDTH: Used in right-aligned printed elements.
    TAB: Default tab-width.
    LOG_LEVEL: Mapping between cli flags and logging flags.
    intro_message: Message displayed during each command.

Note:
    It's assumed that the cli.py will ever only be loaded once per process.
    Like when running it from a terminal, it will be loaded and then the
    entire Python process will be killed.

"""

import os
import sys
import time
import json
import shutil
import logging
import tempfile
import subprocess
import contextlib

from . import api, lib, util, __version__
from .vendor import click

_ctx = None
_help = {
    "main": {
        "paths": "List all available paths",
        "registered-paths": "Print only registered-paths",
        "verbose": "Display detailed information. Useful for "
                   "debugging purposes.",
        "plugin-path": "Replace all normally discovered paths "
                       "with this This may be called multiple times.",
        "add-plugin-path": "Append to normally discovered paths.",
        "logging-level": "Specify with which level to produce "
                         "logging messages. A value lower than the default "
                         "\"warning\" will produce more messages. This "
                         "can be useful for debugging.",
        "environment-paths": "Print only paths added via environment",
        "version": "Print the current version of Pyblish",
        "plugins": "List all available plugins",
        "data": "Initialise context with data. This takes "
                "two arguments, key and value.",
    },
    "publish": {
        "delay": "Add an artificial delay to each plugin. "
                 "Typically used in debugging.",
        "path": "Input path for publishing operation",
        "file": "Load file in host registered to it's suffix",
        "instance": "Only publish specified instance. "
                    "The default behaviour is to publish "
                    "all instances. This may be called multiple times.",
        "targets": "Use only plugins which have similar targets. Provide a "
                   "string of targets separated by a `;`"
    }
}


def _setup_log(root="pyblish"):
    log = logging.getLogger(root)
    log.setLevel(logging.INFO)
    return log


log = _setup_log()
main_log = lib.setup_log(level=logging.ERROR)

PATH_TEMPLATE = "{path} <{typ}>"
LOG_TEMPATE = "{tab}<log>: %(message)s"

SCREEN_WIDTH = 80
TAB = "    "
LOG_LEVEL = {
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warning": logging.WARNING,
    "error": logging.ERROR,
    "critical": logging.CRITICAL
}

intro_message = """pyblish version {version}

Available plugin paths:
{paths}

Available plugins:
{plugins}"""


def _format_paths(paths):
    """Return paths at one new each"""
    message = ""
    for path in paths:
        message += "{0}\n".format(path)
    return message[:-1]  # Discard last newline


def _format_plugins(plugins):
    message = ""
    for plugin in plugins:
        message += "{0}\n".format(plugin.__name__)
    return message[:-1]


def _format_time(start, finish):
    """Return right-aligned time-taken message"""
    message = "Time taken: %.2fs" % (finish - start)
    return message.rjust(SCREEN_WIDTH)


@contextlib.contextmanager
def _cli_plugin(data):
    tempdir = tempfile.mkdtemp()
    fname = os.path.join(tempdir, "cli_plugin.py")
    with open(fname, "w") as f:
        f.write("""\
from pyblish import api

class CliPlugin(api.ContextPlugin):
    \"\"\"Added through command-line interface to modify context\"\"\"
    order = api.CollectorOrder - 0.5
    label = "CLI Plugin"

    def process(self, context):
        import json
        context.data.update({data})
        self.log.debug(context.data)
""".format(data=json.dumps(data)))

    try:
        yield tempdir
    finally:
        shutil.rmtree(tempdir)


@click.group(invoke_without_command=True)
@click.option("--verbose", is_flag=True, help=_help["main"]["verbose"])
@click.option("--version", is_flag=True, help=_help["main"]["version"])
@click.option("--paths", is_flag=True, help=_help["main"]["paths"])
@click.option("--plugins", is_flag=True, help=_help["main"]["plugins"])
@click.option("--registered-paths", is_flag=True,
              help=_help["main"]["registered-paths"])
@click.option("--environment-paths", is_flag=True,
              help=_help["main"]["environment-paths"])
@click.option("-pp",
              "--plugin-path",
              "plugin_paths",
              multiple=True,
              help=_help["main"]["plugin-path"])
@click.option("-ap",
              "--add-plugin-path",
              "add_plugin_paths",
              multiple=True,
              help=_help["main"]["add-plugin-path"])
@click.option("-d",
              "--data",
              nargs=2,
              multiple=True,
              help=_help["main"]["data"])
@click.option("-ll",
              "--logging-level",
              type=click.Choice(LOG_LEVEL.keys()),
              default="error",
              help=_help["main"]["logging-level"])
@click.pass_context
def main(ctx,
         verbose,
         version,
         paths,
         plugins,
         environment_paths,
         registered_paths,
         plugin_paths,
         add_plugin_paths,
         data,
         logging_level):
    """Pyblish command-line interface

    Use the appropriate sub-command to initiate a publish.

    Use the --help flag of each subcommand to learn more
    about what it can do.

    \b
    Usage:
        $ pyblish publish --help
        $ pyblish test --help

    """

    global _ctx
    _ctx = ctx

    # Convert multi-arguments from tuple to list, see #357
    plugin_paths = list(plugin_paths)
    add_plugin_paths = list(add_plugin_paths)
    data = list(data)

    level = LOG_LEVEL[logging_level]
    log.setLevel(level)

    # Process top-level arguments
    if version:
        click.echo("pyblish version %s" % __version__)

    # Respond to sub-commands
    if not ctx.obj:
        ctx.obj = dict()

    # Initialise context with data passed as argument
    context = api.Context()
    ctx.obj["context"] = context

    for key, value in data:
        try:
            value = json.loads(value)
        except ValueError:
            pass

        context.data[str(key)] = value

    if not plugin_paths:
        plugin_paths = api.plugin_paths()
    plugin_paths += add_plugin_paths
    ctx.obj["plugin_paths"] = plugin_paths

    available_plugins = api.discover(paths=plugin_paths)
    if plugins:
        click.echo(_format_plugins(available_plugins))

    if verbose:
        click.echo(
            intro_message.format(
                version=__version__,
                paths=_format_paths(plugin_paths),
                plugins=_format_plugins(available_plugins))
        )

    # Visualise available paths
    if any([paths, environment_paths, registered_paths]):
        _paths = list()

        if paths:
            environment_paths = True
            registered_paths = True

        for path in plugin_paths:

            # Determine the source of each path
            _typ = "custom"
            if path in api.environment_paths():
                _typ = "environment"

            elif path in api.registered_paths():
                _typ = "registered"

            # Only display queried paths
            if _typ == "environment" and not environment_paths:
                continue

            if _typ == "registered" and not registered_paths:
                continue

            click.echo(PATH_TEMPLATE.format(
                path=path, typ=_typ))
            _paths.append(path)

    # Pass data to sub-commands
    ctx.obj["verbose"] = verbose
    ctx.obj["plugin_paths"] = plugin_paths


@click.command()
@click.argument("path", default=".")
@click.option("-i",
              "--instance",
              "instances",
              multiple=True,
              help=_help["publish"]["instance"])
@click.option("-de",
              "--delay",
              default=None,
              type=float,
              help=_help["publish"]["delay"])
@click.option("-t",
              "--targets",
              multiple=True,
              help=_help["publish"]["targets"])
@click.pass_context
def publish(ctx,
            path,
            instances,
            delay,
            targets):
    """Publish instances of path.

    \b
    Arguments:
        path: Optional path, either absolute or relative,
              at which to initialise a publish. Defaults to
              the current working directory.

    \b
    Usage:
        $ pyblish publish my_file.txt --instance=Message01
        $ pyblish publish my_file.txt --all

    """

    # Convert multi-arguments from tuple to list, see #357
    targets = list(targets)
    instances = list(instances)

    _start = time.time()  # Benchmark

    # Use `path` argument as initial data for context
    context = ctx.obj["context"]

    if os.path.isdir(path):
        context.data["current_dir"] = path  # backwards compatibility
        context.data["currentDir"] = path
    else:
        context.data["current_file"] = path  # backwards compatibility
        context.data["currentFile"] = path

    # Begin processing
    plugins = api.discover(paths=ctx.obj["plugin_paths"])
    context = util.publish(context=context, plugins=plugins, targets=targets)

    if any(result["error"] for result in context.data.get("results", [])):
        click.echo("There were errors.")

        for result in context.data["results"]:
            if result["error"] is not None:
                click.echo(result["error"])

    _end = time.time()

    if ctx.obj["verbose"]:
        click.echo()
        click.echo("-" * 80)
        click.echo(_format_time(_start, _end))


@click.command()
@click.argument("package", default="pyblish_qml")
@click.pass_context
def gui(ctx, package):

    environ = os.environ.copy()
    context = ctx.obj["context"]

    registered_guis = api.registered_guis()

    if len(registered_guis) > 0:
        package = registered_guis[0]

    with _cli_plugin(data=context.data) as plugin_path:
        environ["PYBLISHPLUGINPATH"] = os.pathsep.join(
            ctx.obj["plugin_paths"] + [plugin_path]
        )

        process = subprocess.Popen(
            [sys.executable, "-m", package],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            env=environ,
            bufsize=1,
            universal_newlines=True
        )
        while True:
            line = process.stdout.readline()
            if line != '':
                print(line.rstrip())
            else:
                break
        process.wait()
        sys.exit(process.returncode)


main.add_command(publish)
main.add_command(gui)



"""Plug-in system

Works similar to how OSs look for executables; i.e. a number of
absolute paths are searched for a given match. The predicate for
executables is whether or not an extension matches a number of
options, such as ".exe" or ".bat".

In this system, the predicate is whether or not a fname starts
with "validate" and ends with ".py"

"""

# Standard library
import os
import sys
import time
import types
import logging
import inspect
import warnings
import contextlib
import uuid

# Local library
from . import (
    __version__,
    version_info,
    _registered_callbacks,
    _registered_services,
    _registered_plugins,
    _registered_hosts,
    _registered_paths,
    _registered_targets,
    _registered_plugin_filters
)

from . import lib
from .vendor import iscompatible, six

if six.PY2:
    get_arg_spec = inspect.getargspec
else:
    get_arg_spec = inspect.getfullargspec

log = logging.getLogger("pyblish.plugin")

__metaclass__ = type  # Make all classes new-style

# Matching algorithms
Intersection = 1 << 0
Subset = 1 << 1
Exact = 1 << 2

# Check for duplicate plugin names. This is to preserve backwards compatility.
ALLOW_DUPLICATES = bool(os.getenv("PYBLISH_ALLOW_DUPLICATE_PLUGIN_NAMES"))

# Check for strict data types. This is to preserve backwards compatility
STRICT_DATATYPES = bool(os.getenv("PYBLISH_STRICT_DATATYPES"))

# Check for early adopters.
EARLY_ADOPTER = bool(os.getenv("PYBLISH_EARLY_ADOPTER"))
ALLOW_DUPLICATE_PLUGINS = EARLY_ADOPTER or ALLOW_DUPLICATES
STRICT_DATATYPES = EARLY_ADOPTER or STRICT_DATATYPES


class Provider():
    """Dependency provider

    This object is given a series of "services" that it then distributes
    to a passed function based on the function's argument signature.

    For example, the function func:`myfunc(a, b)` is given the services
    called "a" and "b", given they have previously been added to the provider.

    """

    def __init__(self):
        self._services = dict()

    def get(self, service):
        return self.services.get(service)

    @property
    def services(self):
        services = _registered_services.copy()
        services.update(self._services)

        # Forwards-compatibility alias
        services["asset"] = services["instance"]

        return services

    @classmethod
    def args(cls, func):
        return [a for a in get_arg_spec(func)[0]
                if a not in ("self", "cls")]

    def invoke(self, func):
        """Supply function `func` with objects to its signature

        Raises:
            KeyError if an argument asked for is not available

        Returns:
            Result of `func`

        """

        args = self.args(func)
        unavailable = [a for a in args if a not in self.services]

        if unavailable:
            raise KeyError("Unavailable service requested: %s" % unavailable)

        inject = dict((k, v) for k, v in self.services.items()
                      if k in args)

        return func(**inject)

    def inject(self, name, obj):
        self._services[name] = obj


def evaluate_pre11(plugin):
    """Determine whether the plug-in is pre-1.1"""
    plugin.__pre11__ = False

    if hasattr(plugin, "process_context"):
        plugin.__pre11__ = True
        plugin.process = plugin.process_context
        del(plugin.process_context)

    if hasattr(plugin, "process_instance"):
        plugin.__pre11__ = True
        plugin.process = plugin.process_instance
        del(plugin.process_instance)

    # Repair is deprecated
    if hasattr(plugin, "repair_context"):
        plugin.__pre11__ = True
        plugin.repair = plugin.repair_context
        del(plugin.repair_context)

    if hasattr(plugin, "repair_instance"):
        plugin.__pre11__ = True
        plugin.repair = plugin.repair_instance
        del(plugin.repair_instance)


def evaluate_enabledness(plugin):
    """Deterimine whether the plug-in supports Context/Instance"""
    plugin.__contextEnabled__ = False
    plugin.__instanceEnabled__ = False

    args_ = get_arg_spec(plugin.process).args

    if "instance" in args_:
        plugin.__instanceEnabled__ = True

    if "context" in args_:
        plugin.__contextEnabled__ = True

    # Backwards-compatibility with asset
    if "asset" in args_:
        plugin.__instanceEnabled__ = True


def append_logger(plugin):
    """Append logger to plugin

    The logger will include a plug-in's final name, as defined
    by the subclasser. For example, if a plug-in is defined, subclassing
    :class:`Plugin`, it's given name will be present in log records.

    """

    name = plugin.__name__

    # Package name appended, for filtering of LogRecord instances
    logname = "pyblish.%s" % name
    plugin.log = logging.getLogger(logname)
    plugin.log.setLevel(logging.DEBUG)

    # All messages are handled by root-logger
    plugin.log.propagate = True


class MetaPlugin(type):
    """Rewrite plug-ins written prior to 1.1

    ..warning:: In case of plug-ins written prior to 1.1,
        that also process both instance and context,
        only the instance process will remain available.

    """

    def __init__(cls, *args, **kwargs):
        append_logger(cls)
        evaluate_pre11(cls)
        evaluate_enabledness(cls)

        # Compute once
        cls._id = str(uuid.uuid4())
        cls.id = lib.classproperty(lambda self: self._id)

        return super(MetaPlugin, cls).__init__(*args, **kwargs)


@lib.log
@six.add_metaclass(MetaPlugin)
class Plugin():
    """Base-class for plugins

    Attributes:
        hosts: Optionally limit a plug-in to one or more hosts
        families: Optionally limit a plug-in to one or more families
        label: Printed name of plug-in
        active: Whether or not to use plug-in during processing
        version: Optional version for forwards-compatibility.
            Pyblish is (currently not) using the version to allow
            for plug-ins incompatible with a particular running
            instance of Pyblish to co-exist alongside compatible
            versions.
        order: Order in which this plug-in is processed. This is
            used internally to control which plug-ins are processed
            before another so as to allow plug-ins to communicate
            with each other. E.g. one plug-in may provide critical
            information to another and so must be allowed to be
            processed first.
        optional: Whether or not plug-in can be skipped by the user.
        requires: Which version of Pyblish is required by this plug-in.
            Plug-ins requiring a version newer than the current version
            will not be loaded. 1.0.8 was when :attr:`Plugin.requires`
            was first introduced.
        actions: Actions associated to this plug-in
        id: Unique ID as str
        match: Family matching algorithm - Intersection, Subset or Exact
            Intersection -> set(a).intersection(b)
            Subset       -> set(a).issubset(b)
            Exact        -> a == b

    """

    hosts = ["*"]
    families = ["*"]
    targets = ["default"]
    label = None
    active = True
    version = (0, 0, 0)
    order = -1
    optional = False
    requires = "pyblish>=1"
    actions = []
    id = None  # Defined by metaclass
    match = Intersection  # Default matching algorithm

    def __str__(self):
        return self.label or type(self).__name__

    def __repr__(self):
        return u"%s.%s(%r)" % (__name__, type(self).__name__, self.__str__())

    def process(self):
        """Primary processing method

        This method is called whenever your plug-in is invoked
        and is injected with object relative to it's signature.

        E.g. process(self, context, instance) will have the current
        context and instance injected into it at run-time.

        Available objects:
            - context
            - instance
            - user
            - time

        Raises:
            Any error

        """

        pass

    def repair(self):
        """DEPRECATED"""
        pass


class Collector(Plugin):
    """Parse a given working scene for available Instances"""

    order = 0


class Validator(Plugin):
    """Validate/check/test individual instance for correctness."""

    order = 1


class Extractor(Plugin):
    """Physically separate Instance from Host into corresponding resources"""

    order = 2


class Integrator(Plugin):
    """Integrates publishes into a pipeline"""

    order = 3


CollectorOrder = 0
ValidatorOrder = 1
ExtractorOrder = 2
IntegratorOrder = 3


def validate_argument_signature(plugin):
    """Ensure plug-in processes either 'instance' or 'context'"""
    if not any(arg in get_arg_spec(plugin.process).args
               for arg in ("instance", "context")):
        plugin.__invalidSignature__ = True


class ExplicitMetaPlugin(MetaPlugin):
    """Validate explicit plug-ins"""

    def __init__(cls, *args, **kwargs):
        validate_argument_signature(cls)
        return super(ExplicitMetaPlugin, cls).__init__(*args, **kwargs)


@six.add_metaclass(ExplicitMetaPlugin)
class ContextPlugin(Plugin):

    def process(self, context):
        """Primary processing method

        Arguments:
            context (Context): Context with which to process

        """


@six.add_metaclass(MetaPlugin)
class InstancePlugin(Plugin):

    def process(self, instance):
        """Primary processing method

        Arguments:
            instance (Instance): Instance with which to process

        """


class MetaAction(type):
    """Inject additional metadata into Action"""

    def __init__(cls, *args, **kwargs):
        cls._id = str(uuid.uuid4())
        cls.id = lib.classproperty(lambda self: cls._id)

        cls.__error__ = None

        if cls.on not in ("all",
                          "notProcessed",
                          "processed",
                          "failed",
                          "warning",
                          "failedOrWarning",
                          "succeeded"):
            cls.__error__ = (
                "Action had an unrecognised value "
                "for `on`: \"%s\"" % cls.on
            )

        return super(MetaAction, cls).__init__(*args, **kwargs)


@lib.log
@six.add_metaclass(MetaAction)
class Action():
    """User-supplied interactive action

    Subclass this class and append to Plugin.actions in order
    to provide your users with optional, context sensitive
    functionality.

    Attributes:
        label: Optional label to display in place of class name.
        active: Whether or not to allow execution of action.
        on: When to enable this action; available options are:
            - "all": Always available (default).
            - "notProcessed": The plug-in has not yet been processed
            - "processed": The plug-in has been processed
            - "succeeded": The plug-in has been processed, and succeeded
            - "failed": The plug-in has been processed, and failed
            - "warning": The plug-in has been processed, and had a warning
            - "failedOrWarning": The plug-in has been processed, and failed or
              had a warning
        icon: Name, relative path or absolute path to image for
            use as an icon of this action. For relative paths,
            the current working directory of the host is used and
            names represent icons available via Awesome Icons.
            fortawesome.github.io/Font-Awesome/icons/

    """

    __type__ = "action"

    label = None
    active = True
    on = "all"
    icon = None

    def __str__(self):
        return self.label or type(self).__name__

    def __repr__(self):
        return u"%s.%s(%r)" % (__name__, type(self).__name__, self.__str__())

    def process(self):
        pass


class Separator(Action):
    __type__ = "separator"


def Category(label):
    return type("Category", (Action,), {"label": label,
                                        "__type__": "category"})


@contextlib.contextmanager
def logger(handler):
    """Listen in on the global logger

    Arguments:
        handler (Handler): Custom handler with which to use
            to listen for log messages

    """

    logger = logging.getLogger()
    old_level = logger.level

    logger.addHandler(handler)
    logger.setLevel(logging.DEBUG)

    try:
        yield
    finally:
        logger.removeHandler(handler)
        logger.setLevel(old_level)


def process(plugin, context, instance=None, action=None):
    """Produce a single result from a Plug-in

    Arguments:
        plugin(Plugin): Uninstantiated plug-in class
        context(Context): The current Context
        instance(Instance, optional): Instance to process
        action(str): Id of action to process, in place of plug-in.

    Returns:
        Dictionary of result

    """

    if issubclass(plugin, (ContextPlugin, InstancePlugin)):
        result = __explicit_process(plugin, context, instance, action)
    else:
        result = __implicit_process(plugin, context, instance, action)

    lib.emit("pluginProcessed", result=result)
    return result


def __explicit_process(plugin, context, instance=None, action=None):
    """Produce result from explicit plug-in

    This is the primary internal mechanism for producing results
    from the processing of plug-in/instance pairs.

    This mechanism replaces :func:`__implicit_process`.

    """

    if not action and issubclass(plugin, InstancePlugin) and instance is None:
        raise AssertionError("Cannot process an InstancePlugin without an "
                             "instance. This is a bug")

    result = {
        "success": False,
        "plugin": plugin,
        "instance": instance,
        "action": action,
        "error": None,
        "records": list(),
        "duration": None,
        "progress": 0,
    }

    if not action:
        args = (context if issubclass(plugin, ContextPlugin) else instance,)
        runner = plugin().process
    else:
        actions = dict((a.id, a) for a in plugin.actions)
        assert action in actions, ("%s did not have action: %s. This is a bug"
                                   % (plugin, action))
        action = actions[action]
        args = (context, plugin)
        runner = action().process

    records = list()
    handler = lib.MessageHandler(records)

    __start = time.time()

    try:
        with logger(handler):
            runner(*args)
            result["success"] = True
    except Exception as error:
        # FIXME: This is apparently not very healthy,
        # as it creates a circular reference.
        # http://stackoverflow.com/a/11417308/478949
        lib.emit("pluginFailed", plugin=plugin, context=context,
                 instance=instance, error=error)
        lib.extract_traceback(error, plugin.__module__)
        result["error"] = error
        log.exception(result["error"].formatted_traceback)

    __end = time.time()

    for record in records:
        result["records"].append(record)

    result["duration"] = (__end - __start) * 1000  # ms

    if "results" not in context.data:
        context.data["results"] = list()

    context.data["results"].append(result)

    return result


def __implicit_process(plugin, context, instance=None, action=None):
    """Produce result from implicit plug-in

    This is a fallback mechanism for backwards compatibility.
    An implicit plug-in are those subclassed from Collector,
    Validator, Extractor or Integrator.

    The mechanism which replaces this is :func:`__explicit_process`.

    """

    result = {
        "success": False,
        "plugin": plugin,
        "instance": instance,
        "action": action,
        "error": None,
        "records": list(),
        "duration": None,
        "progress": 0,
    }

    if not action:
        runner = plugin().process
    else:
        actions = dict((a.id, a) for a in plugin.actions)
        assert action in actions, ("%s did not have action: %s. This is a bug"
                                   % (plugin, action))
        action = actions[action]
        runner = action().process

    records = list()
    handler = lib.MessageHandler(records)

    provider = Provider()
    provider.inject("plugin", plugin)
    provider.inject("context", context)
    provider.inject("instance", instance)

    __start = time.time()

    try:
        with logger(handler):
            provider.invoke(runner)
            result["success"] = True
    except Exception as error:
        lib.emit("pluginFailed", plugin=plugin, context=context,
                 instance=instance, error=error)
        lib.extract_traceback(error, plugin.__module__)
        result["error"] = error
        log.exception(result["error"].formatted_traceback)

    __end = time.time()

    for record in records:
        result["records"].append(record)

    result["duration"] = (__end - __start) * 1000  # ms

    if "results" not in context.data:
        context.data["results"] = list()

    context.data["results"].append(result)

    # Backwards compatibility
    result["asset"] = instance  # Deprecated key

    return result


def repair(plugin, context, instance=None):
    """Produce single result from repairing"""

    import time

    if "results" not in context.data:
        context.data["results"] = list()

    result = {
        "success": False,
        "plugin": plugin,
        "instance": instance,
        "error": None,
        "records": list(),
        "duration": None
    }

    plugin = plugin()

    records = list()
    handler = lib.MessageHandler(records)

    provider = Provider()
    provider.inject("context", context)
    provider.inject("instance", instance)

    __start = time.time()

    try:
        with logger(handler):
            provider.invoke(plugin.repair)
            result["success"] = True
    except Exception as error:
        lib.extract_traceback(error)
        result["error"] = error

    __end = time.time()

    for record in records:
        result["records"].append(record)

    result["duration"] = (__end - __start) * 1000  # ms

    context.data["results"].append(result)

    return result


class _Dict(dict):
    """Temporary object during transition from set_data to data dictionary"""

    def __init__(self, parent):
        self._parent = parent

    def __call__(self, key=None, default=None):
        if key is None:
            return self.copy()

        if key == "name":
            default = self._parent.name

        return self.get(key, default)

    def __setitem__(self, k, v):
        # Backwards incompatible data validation.
        if STRICT_DATATYPES:
            # Validate "publish" data member to always be boolean
            if k == "publish" and not isinstance(v, bool):
                raise TypeError("\"publish\" data member has to be boolean.")

        dict.__setitem__(self, k, v)


class AbstractEntity(list):
    """Superclass for Context and Instance

    Attributes:
        id (str): Unique identifier of instance
        name (str): Name of instance
        data (dict): Data shared between plug-ins
        parent (AbstractEntity): Optional parent of instance

    """

    def __init__(self, name, parent=None):
        assert isinstance(name, six.string_types)
        assert parent is None or isinstance(parent, AbstractEntity)

        # Read-only properties
        self._name = name
        self._data = _Dict(self)
        self._id = str(uuid.uuid4())
        self._parent = parent

        if parent is not None:
            parent.append(self)

    @property
    def id(self):
        return self._id

    @property
    def parent(self):
        return self._parent

    @property
    def name(self):
        return self._name

    @property
    def data(self):
        return self._data


class Context(AbstractEntity):
    """Maintain a collection of Instances"""

    def __init__(self, name="Context", parent=None):
        super(Context, self).__init__(name, parent)

    def __contains__(self, key):
        """Support both Instance objects and `id` strings

        Example:
            >>> context = Context()
            >>> instance = context.create_instance("MyInstance")
            >>> instance.id in context
            True
            >>> instance in context
            True
            >>> "NotExists" in context
            False

        """

        try:
            key = key.id
        except Exception:
            pass

        return key in [c.id for c in self]

    def create_instance(self, name, **kwargs):
        """Convenience method of the following.

        >>> ctx = Context()
        >>> inst = Instance("name", parent=ctx)

        Example:
            >>> ctx = Context()
            >>> inst = ctx.create_instance(name="Name")

        """

        instance = Instance(name, parent=self, **kwargs)
        return instance

    def __getitem__(self, item):
        """Enable support for dict-like getting of children by id

        Example:
            >>> context = Context()
            >>> instance = context.create_instance("MyInstance")
            >>> assert context[instance.id].name == "MyInstance"
            >>> assert context[0].name == "MyInstance"

        """

        if isinstance(item, int):
            return super(Context, self).__getitem__(item)
        try:
            return next(c for c in self if c.id == item)
        except StopIteration:
            raise KeyError("%s not in list" % item)

    def get(self, key, default=None):
        """Enable support for dict-like getting of children by id

        Example
            >>> context = Context()
            >>> instance = context.create_instance("MyInstance")
            >>> assert context.get(instance.id).name == "MyInstance"

        """

        return next((c for c in self if c.id == key), default)


@lib.log
class Instance(AbstractEntity):
    """An in-memory representation of one or more files

    Examples include rigs, models.

    Arguments:
        name (str): Name of instance, typically used during
            extraction as name of resulting files.
        parent (AbstractEntity): Optional parent. This is
            supplied automatically when creating instances with
            :class:`Context.create_instance()`.

    """

    def __init__(self, name, parent=None, **kwargs):
        super(Instance, self).__init__(name, parent)
        self._data["family"] = "default"
        self._data["name"] = name
        self._data.update(kwargs)

    def __eq__(self, other):
        return self._id == getattr(other, "id", None)

    def __ne__(self, other):
        return self._id != getattr(other, "id", None)

    def __repr__(self):
        return u"%s.%s(\"%s\")" % (__name__, type(self).__name__, self)

    def __str__(self):
        return self._name

    @property
    def context(self):
        """Return top-level parent; the context"""
        parent = self.parent
        while parent.parent:
            try:
                parent = parent.parent
            except Exception:
                break

        assert isinstance(parent, Context), ("Parent was not a Context:"
                                             "%s" % type(parent))

        return parent


# Forwards-compatibility alias
Asset = Instance


def current_host():
    """Return host last registered thru `register_host()`

    When running Pyblish from within a host, this function determines
    which host is running and returns the equivalent keyword.

    Example:
        >> # Running within Autodesk Maya
        >> current_host()
        "maya"
        >> # Running within Sidefx Houdini
        >> current_host()
        "houdini"
        >> # Running from an unknown software
        >> current_host()
        "unknown"

    """

    return _registered_hosts[-1] if _registered_hosts else "unknown"


def register_callback(signal, callback):
    """Register a new callback

    Arguments:
        signal (string): Name of signal to register the callback with.
        callback (func): Function to execute when a signal is emitted.

    Raises:
        ValueError if `callback` is not callable.

    """

    if not hasattr(callback, "__call__"):
        raise ValueError("%s is not callable" % callback)

    if signal in _registered_callbacks:
        _registered_callbacks[signal].append(callback)
    else:
        _registered_callbacks[signal] = [callback]


def deregister_callback(signal, callback):
    """Deregister a callback

    Arguments:
        signal (string): Name of signal to deregister the callback with.
        callback (func): Function to execute when a signal is emitted.

    Raises:
        KeyError on missing signal
        ValueError on missing callback
    """

    _registered_callbacks[signal].remove(callback)


def deregister_all_callbacks():
    """Deregisters all callback"""

    _registered_callbacks.clear()


def registered_callbacks():
    """Returns registered callbacks"""

    return _registered_callbacks


def register_plugin(plugin):
    """Register a new plug-in

    Arguments:
        plugin (Plugin): Plug-in to register

    Raises:
        TypeError if `plugin` is not callable

    """

    if not hasattr(plugin, "__call__"):
        raise TypeError("Plug-in must be callable "
                        "returning an instance of a class")

    if not plugin_is_valid(plugin):
        raise TypeError("Plug-in invalid: %s", plugin)

    if not version_is_compatible(plugin):
        raise TypeError(
            "Plug-in %s not compatible with "
            "this version (%s) of Pyblish." % (
                plugin, __version__))

    if not host_is_compatible(plugin):

        hosts = registered_hosts()
        required_hosts = plugin.hosts

        err = """Plug-in %s is not compatible with available host(s).

Required host(s): %s
Registered host(s): %s

Make sure the integration for your host is correctly setup
or register a new host using `pyblish.api.register_host("%s")`
""" % (plugin, repr(required_hosts), repr(hosts), required_hosts[0])

        raise TypeError(err)

    _registered_plugins[plugin.__name__] = plugin


def deregister_plugin(plugin):
    """De-register an existing plug-in

    Arguments:
        plugin (Plugin): Existing plug-in to de-register

    """

    _registered_plugins.pop(plugin.__name__)


def deregister_all_plugins():
    """De-register all plug-ins"""
    _registered_plugins.clear()


@lib.deprecated
def register_service(name, obj):
    """Register a new service

    Arguments:
        name (str): Name of service
        obj (any): Any object

    """

    _registered_services[name] = obj


@lib.deprecated
def deregister_service(name):
    """De-register an existing service by name

    Arguments:
        name (str): Name of service

    """

    _registered_services.pop(name)


@lib.deprecated
def deregister_all_services():
    """De-register all existing services"""
    _registered_services.clear()


@lib.deprecated
def registered_services():
    """Return the currently registered services as a dictionary

    .. note:: This returns a copy of the registered paths
        and can therefore not be modified directly.

    """

    return _registered_services.copy()


def register_plugin_path(path):
    """Plug-ins are looked up at run-time from directories registered here

    To register a new directory, run this command along with the absolute
    path to where you"re plug-ins are located.

    Example:
        >>> import os
        >>> my_plugins = os.path.join("server", "plugins")
        >>> register_plugin_path(my_plugins) == os.path.normpath(my_plugins)
        True

    Returns:
        Actual path added, including any post-processing

    """

    # accept pathlib paths and convert to string
    if hasattr(path, "as_posix"):
        path = path.as_posix()

    normpath = os.path.normpath(path)
    if normpath in _registered_paths:
        return log.warning("Path already registered: {0}".format(path))

    _registered_paths.append(normpath)

    return path


def deregister_plugin_path(path):
    """Remove a pyblish._registered_paths path

    Raises:
        ValueError if `path` isn't registered

    """

    # accept pathlib paths and convert to string
    if hasattr(path, "as_posix"):
        path = path.as_posix()

    normpath = os.path.normpath(path)
    try:
        _registered_paths.remove(normpath)
    except ValueError:
        return log.error("Path isn't registered: {0}".format(path))


def deregister_all_paths():
    """Mainly used in tests"""
    _registered_paths[:] = []


def registered_paths():
    """Return paths added via registration

    ..note:: This returns a copy of the registered paths
        and can therefore not be modified directly.

    """

    return list(_registered_paths)


def registered_plugins():
    """Return plug-ins added via :func:`register_plugin`

    .. note:: This returns a copy of the registered plug-ins
        and can therefore not be modified directly

    """

    plugins = list()

    for plugin in _registered_plugins.values():
        # Maintain immutability across retrievals
        copy = type(plugin.__name__, (plugin,), {})
        copy._id = plugin._id
        copy.__doc__ = plugin.__doc__
        plugins.append(copy)

    return plugins


def register_host(host):
    """Register a new host

    Registered hosts are used to filter discovered
    plug-ins by host.

    Example:
        >>> register_host("My Host")
        >>> "My Host" in registered_hosts()
        True

    """

    if host not in _registered_hosts:
        _registered_hosts.append(host)


def deregister_host(host, quiet=False):
    """Remove an already registered host

    Arguments:
        host (str): Name of host
        quiet (bool): Whether to raise an exception
            when attempting to remove a host that is
            not already registered.

    """

    try:
        _registered_hosts.remove(host)
    except Exception as e:
        if not quiet:
            raise e


def deregister_all_hosts():
    _registered_hosts[:] = []


def registered_hosts():
    """Return the currently registered hosts"""
    return list(_registered_hosts)


def current_target():
    return _registered_targets[-1] if _registered_targets else ""


def register_target(target):
    """Register a new target

    Registered targets can be used in plug-ins to determin outputs

    Example:
        >>> register_target("Studio")
        >>> "Studio" in registered_targets()
        True
        >>> current_target()
        'Studio'

    """

    if target in _registered_targets:
        idx = _registered_targets.index(target)
        _registered_targets.pop(idx)

    _registered_targets.append(target)


def deregister_target(target, quiet=False):
    """Remove an already registered target

    Arguments:
        target (str): Name of target
        quiet (bool): Whether to raise an exception
            when attempting to remove a target that is
            not already registered.

    """

    try:
        _registered_targets.remove(target)
    except Exception as e:
        if not quiet:
            raise e


def deregister_all_targets():
    _registered_targets[:] = []


def registered_targets():
    """Return the currently registered targets"""
    return list(_registered_targets)


def register_discovery_filter(callback):
    """Register a new plugin filter

    Arguments:
        callback (func): Function to execute on filter during discovery,
            takes the original of plugins to be edited in-place

    Raises:
        ValueError if `callback` is not callable.

    """

    if not callable(callback):
        raise ValueError("%s is not callable" % callback)

    _registered_plugin_filters.append(callback)


def deregister_discovery_filter(callback):
    """Deregister a plugin filter

    Arguments:
        callback (func): filtering function.

    Raises:
        ValueError on missing callback
    """

    _registered_plugin_filters.remove(callback)


def deregister_all_discovery_filters():
    """Deregisters all plugin filters"""
    _registered_plugin_filters[:] = []


def registered_discovery_filters():
    """Returns registered plugin filter callbacks"""

    return _registered_plugin_filters


def environment_paths():
    """Return paths added via environment variable"""

    plugin_path = os.environ.get("PYBLISHPLUGINPATH")
    if not plugin_path:
        return list()

    paths = plugin_path.split(os.pathsep)
    log.debug("Paths from environment: %s" % paths)

    return paths


def plugin_paths():
    """Collect paths from all sources.

    This function looks at the three potential sources of paths
    and returns a list with all of them together.

    The sources are:

    - Registered paths using :func:`register_plugin_path`,
    - Paths from the environment variable `PYBLISHPLUGINPATH`

    Returns:
        list of paths in which plugins may be locat

    """

    paths = list()

    for path in registered_paths() + environment_paths():
        if path in paths:
            continue
        paths.append(path)

    return paths


def discover(type=None, regex=None, paths=None):
    """Find and return available plug-ins

    This function looks for files within paths registered via
    :func:`register_plugin_path` and those added to `PYBLISHPLUGINPATH`.

    It determines *type* - :class:`Selector`, :class:`Validator`,
    :class:`Extractor` or :class:`Conform` - based on whether it
    matches it's corresponding regular expression; e.g.
    "$validator_*^" for plug-ins of type Validator.

    Arguments:
        type (str, optional): !DEPRECATED! Only return plugins of
            specified type. E.g. validators, extractors. In None is specified,
            return all plugins. Available options are "selectors", validators",
            "extractors", "conformers", "collectors" and "integrators".
        regex (str, optional): Limit results to those matching `regex`.
            Matching is done on classes, as opposed to
            filenames, due to a file possibly hosting
            multiple plugins.
        paths (list, optional): Paths to discover plug-ins from.
            If no paths are provided, all paths are searched.

    """

    if type is not None:
        warnings.warn("type argument has been deprecated and does nothing")

    if regex is not None:
        warnings.warn("discover(): regex argument "
                      "has been deprecated and does nothing")

    plugins = dict()
    plugin_names = []

    # Include plug-ins from registered paths
    for path in paths or plugin_paths():
        path = os.path.normpath(path)
        if not os.path.isdir(path):
            log.debug("Skipped: \"%s\", path is not a valid folder", path)
            continue

        for fname in os.listdir(path):
            if fname.startswith("_"):
                log.debug("Skipped: \"%s\", starts with _", fname)
                continue

            abspath = os.path.join(path, fname)

            if not os.path.isfile(abspath):
                log.debug("Skipped: \"%s\", not a valid file", abspath)
                continue

            mod_name, mod_ext = os.path.splitext(fname)

            if not mod_ext == ".py":
                log.debug("Skipped: \"%s\",\"%s\", not end in .py", mod_name, mod_ext)
                continue

            module = types.ModuleType(mod_name)
            module.__file__ = abspath

            try:
                with open(abspath, "rb") as f:
                    six.exec_(f.read(), module.__dict__)

                # Store reference to original module, to avoid
                # garbage collection from collecting it's global
                # imports, such as `import os`.
                sys.modules[abspath] = module

            except Exception as err:
                log.error("Skipped: \"%s\" (%s)", mod_name, err)
                continue

            for plugin in plugins_from_module(module):
                if not ALLOW_DUPLICATES and plugin.__name__ in plugin_names:
                    log.debug("Duplicate plug-in found: %s", plugin)
                    continue

                plugin_names.append(plugin.__name__)

                plugin.__module__ = module.__file__
                key = "{0}.{1}".format(plugin.__module__, plugin.__name__)
                plugins[key] = plugin

    # Include plug-ins from registration.
    # Directly registered plug-ins take precedence.
    for plugin in registered_plugins():
        if not ALLOW_DUPLICATES and plugin.__name__ in plugin_names:
            log.debug("Duplicate plug-in found: %s", plugin)
            continue

        plugin_names.append(plugin.__name__)

        plugins[plugin.__name__] = plugin

    plugins = list(plugins.values())
    sort(plugins)  # In-place

    # In-place user-defined filter
    for filter_ in _registered_plugin_filters:
        filter_(plugins)

    return plugins


def plugins_from_module(module):
    """Return plug-ins from module

    Arguments:
        module (types.ModuleType): Imported module from which to
            parse valid Pyblish plug-ins.

    Returns:
        List of plug-ins, or empty list if none is found.

    """

    plugins = list()

    for name in dir(module):
        if name.startswith("_"):
            continue

        # It could be anything at this point
        obj = getattr(module, name)

        if not inspect.isclass(obj):
            continue

        if not issubclass(obj, Plugin):
            continue

        if not plugin_is_valid(obj):
            log.debug("Plug-in invalid: %s", obj)
            continue

        if not version_is_compatible(obj):
            log.debug("Plug-in %s not compatible with "
                      "this version (%s) of Pyblish." % (
                          obj, __version__))
            continue

        if not host_is_compatible(obj):
            log.debug("No supported host found for plugin:%s",  obj)
            continue

        plugins.append(obj)

    return plugins


def plugin_is_valid(plugin):
    """Determine whether or not plug-in `plugin` is valid

    Arguments:
        plugin (Plugin): Plug-in to assess

    """

    if not isinstance(plugin.requires, six.string_types):
        log.debug("Plug-in requires must be of type string: %s", plugin)
        return False

    if not isinstance(plugin.families, list):
        log.debug(".families must be list of stirngs")
        return False

    if not isinstance(plugin.targets, list):
        log.debug(".targets must be list of strings")
        return False

    if not isinstance(plugin.hosts, list):
        log.debug(".hosts must be list of strings")
        return False

    for family in plugin.families:
        if not isinstance(family, six.string_types):
            log.debug("Families must be string")
            return False

    for host in plugin.hosts:
        if not isinstance(host, six.string_types):
            log.debug("Hosts must be string")
            return False

    if hasattr(plugin, "__invalidSignature__"):
        log.debug("Invalid signature")
        return False

    if plugin.match not in (Intersection, Subset, Exact):
        log.debug("'%s' not a supported family "
                  "matching algorithm." % plugin.match)
        log.debug("Options are "
                  "pyblish.api.Intersection, "
                  "pyblish.api.Subset and"
                  "pyblish.api.Exact")
        return False

    return True


def version_is_compatible(plugin):
    """Lookup compatibility between plug-in and current version of Pyblish

    Arguments:
        plugin (Plugin): Plug-in to test against

    """

    if not iscompatible.iscompatible(requirements=plugin.requires,
                                     version=version_info):
        return False
    return True


def host_is_compatible(plugin):
    """Determine whether plug-in `plugin` is compatible with the current host

    Available hosts are determined by :func:`registered_hosts`.

    Arguments:
        plugin (Plugin): Plug-in to assess.

    """

    if "*" in plugin.hosts:
        return True

    return any(host in plugin.hosts for host in registered_hosts())


def sort(plugins):
    """Sort `plugins` in-place

    Their order is determined by their `order` attribute,
    which defaults to their standard execution order:

        1. Selection
        2. Validation
        3. Extraction
        4. Conform

    *But may be overridden.

    Arguments:
        plugins (list): Plug-ins to sort

    """

    if not isinstance(plugins, list):
        raise TypeError("plugins must be of type list")

    plugins.sort(key=lambda p: p.order)
    return plugins




# Standard library
import os
import random

# Local library
import pyblish.plugin

from .lib import (
    setup,
    teardown,
    setup_duplicate,
    setup_empty
)
from nose.tools import (
    with_setup,
    assert_equals,
    assert_true,
    assert_raises
)


@with_setup(setup, teardown)
def test_print_plugin():
    """Printing plugin returns name of class"""
    plugins = pyblish.plugin.discover('validators')
    plugin = plugins[0]
    assert plugin.__name__ in repr(plugin())
    assert plugin.__name__ == str(plugin())


@with_setup(setup, teardown)
def test_name_override():
    """Instances return either a data-member of name or its native name"""
    inst = pyblish.plugin.Instance(name='my_name')
    assert inst.data('name') == 'my_name'

    inst.set_data('name', value='overridden_name')
    assert inst.data('name') == 'overridden_name'


@with_setup(setup_duplicate, teardown)
def test_no_duplicate_plugins():
    """Discovering plugins results in a single occurence of each plugin"""
    plugin_paths = pyblish.plugin.plugin_paths()
    assert_equals(len(plugin_paths), 2)

    plugins = pyblish.plugin.discover(type='selectors')

    # There are two plugins available, but one of them is
    # hidden under the duplicate module name. As a result,
    # only one of them is returned. A log message is printed
    # to alert the user.
    assert_equals(len(plugins), 1)


def test_entities_prints_nicely():
    """Entities Context and Instance prints nicely"""
    ctx = pyblish.plugin.Context()
    inst = ctx.create_instance(name='Test')
    assert 'Instance' in repr(inst)
    assert 'pyblish.plugin' in repr(inst)


def test_deregister_path():
    path = "/server/plugins"
    pyblish.plugin.register_plugin_path(path)
    assert os.path.normpath(path) in pyblish.plugin.registered_paths()
    pyblish.plugin.deregister_plugin_path(path)
    assert os.path.normpath(path) not in pyblish.plugin.registered_paths()


def test_environment_paths():
    """Registering via the environment works"""
    key = "PYBLISHPLUGINPATH"
    path = '/test/path'
    existing = os.environ.get(key)

    try:
        os.environ[key] = path
        assert path in pyblish.plugin.plugin_paths()
    finally:
        os.environ[key] = existing or ''


def test_instances_by_plugin_invariant():
    ctx = pyblish.plugin.Context()
    for i in range(10):
        inst = ctx.create_instance(name="Instance%i" % i)
        inst.set_data("family", "A")

        if i % 2:
            # Every other instance is of another family
            inst.set_data("family", "B")

    class MyPlugin(pyblish.plugin.Validator):
        hosts = ["python"]
        families = ["A"]

        def process(self, instance):
            pass

    compatible = pyblish.logic.instances_by_plugin(ctx, MyPlugin)

    # Test invariant
    #
    # in:  [1, 2, 3, 4]
    # out: [1, 4] --> good
    #
    # in:  [1, 2, 3, 4]
    # out: [2, 1, 4] --> bad
    #

    def test():
        for instance in compatible:
            assert ctx.index(instance) >= compatible.index(instance)

    test()

    compatible.reverse()
    assert_raises(AssertionError, test)


def test_plugins_by_family_wildcard():
    """Plug-ins with wildcard family is included in every query"""
    Plugin1 = type("Plugin1",
                   (pyblish.api.Validator,),
                   {"families": ["myFamily"]})
    Plugin2 = type("Plugin2",
                   (pyblish.api.Validator,),
                   {"families": ["*"]})

    assert Plugin2 in pyblish.api.plugins_by_family(
        [Plugin1, Plugin2], "myFamily")


@with_setup(setup, teardown)
def test_plugins_sorted():
    """Plug-ins are returned sorted by their `order` attribute"""
    plugins = pyblish.api.discover()
    random.shuffle(plugins)  # Randomise their order
    pyblish.api.sort_plugins(plugins)

    order = 0
    for plugin in plugins:
        assert_true(plugin.order >= order)
        order = plugin.order

    assert order > 0, plugins


@with_setup(setup_empty, teardown)
def test_inmemory_plugins():
    """In-memory plug-ins works fine"""

    class InMemoryPlugin(pyblish.api.Selector):
        hosts = ["*"]
        families = ["*"]

        def process_context(self, context):
            context.set_data("workingFine", True)

    pyblish.api.register_plugin(InMemoryPlugin)

    context = pyblish.api.Context()
    for result in pyblish.logic.process(
            func=pyblish.plugin.process,
            plugins=pyblish.api.discover,
            context=context):
        assert_true(result["plugin"].id == InMemoryPlugin.id)

    assert context.data("workingFine") is True


@with_setup(setup_empty, teardown)
def test_inmemory_query():
    """Asking for registered plug-ins works well"""

    InMemoryPlugin = type("InMemoryPlugin", (pyblish.api.Selector,), {})
    pyblish.api.register_plugin(InMemoryPlugin)
    assert pyblish.api.registered_plugins()[0].id == InMemoryPlugin.id


@with_setup(setup_empty, teardown)
def test_plugin_families_defaults():
    """Plug-ins without specific families default to wildcard"""

    class SelectInstances(pyblish.api.Selector):
        def process(self, instance):
            pass

    instance = pyblish.api.Instance("MyInstance")
    instance.set_data("family", "SomeFamily")

    assert_equals(pyblish.api.instances_by_plugin(
        [instance], SelectInstances)[0], instance)

    class ValidateInstances(pyblish.api.Validator):
        def process(self, instance):
            pass

    assert_equals(pyblish.api.instances_by_plugin(
        [instance], ValidateInstances)[0], instance)



"""Compatibility module"""

import re
import inspect
import warnings
from . import plugin, lib, logic
from .vendor import six

if six.PY2:
    get_arg_spec = inspect.getargspec
else:
    get_arg_spec = inspect.getfullargspec

# Aliases
Selector = plugin.Collector
Conformer = plugin.Integrator

_filename_ascii_strip_re = re.compile(r'[^-\w.]')
_windows_device_files = ('CON', 'AUX', 'COM1', 'COM2', 'COM3', 'COM4',
                         'LPT1', 'LPT2', 'LPT3', 'PRN', 'NUL')


def sort(*args, **kwargs):
    warnings.warn("pyblish.api.sort has been deprecated; "
                  "use pyblish.api.sort_plugins")
    return plugin.sort(*args, **kwargs)


def deregister_all(*args, **kwargs):
    warnings.warn("pyblish.api.deregister_all has been deprecated; "
                  "use pyblish.api.deregister_all_paths")
    return plugin.deregister_all_paths(*args, **kwargs)


# AbstractEntity
#
# The below members represent backwards compatibility
# features, kept separate for maintainability as they
# will no longer be updated and eventually discarded.


@lib.deprecated
def set_data(self, key, value):
    """DEPRECATED - USE .data DICTIONARY DIRECTLY

    Modify/insert data into entity

    Arguments:
        key (str): Name of data to add
        value (object): Value of data to add

    """

    self.data[key] = value


@lib.deprecated
def remove_data(self, key):
    """DEPRECATED - USE .data DICTIONARY DIRECTLY

    Remove data from entity

    Arguments;
        key (str): Name of data to remove

    """

    self.data.pop(key)


@lib.deprecated
def has_data(self, key):
    """DEPRECATED - USE .data DICTIONARY DIRECTLY

    Check if entity has key

    Arguments:
        key (str): Key to check

    Return:
        True if it exists, False otherwise

    """

    return key in self.data


@lib.deprecated
def add(self, other):
    """DEPRECATED - USE .append

    Add member to self

    This is to mimic the interface of set()

    """

    return self.append(other)


@lib.deprecated
def remove(self, other):
    """DEPRECATED - USE .pop

    Remove member from self

    This is to mimic the interface of set()

    """

    index = self.index(other)
    return self.pop(index)


plugin.AbstractEntity.add = add
plugin.AbstractEntity.remove = remove
plugin.AbstractEntity.set_data = set_data
plugin.AbstractEntity.remove_data = remove_data
plugin.AbstractEntity.has_data = has_data


# Context

@lib.deprecated
def create_asset(self, *args, **kwargs):
    return self.create_instance(*args, **kwargs)


@lib.deprecated
def add(self, other):
    return super(plugin.Context, self).append(other)


plugin.Context.create_asset = create_asset
plugin.Context.add = add


@lib.deprecated
def format_filename(filename):
    return filename


@lib.deprecated
def format_filename2(filename):
    return filename


lib.format_filename = format_filename
lib.format_filename2 = format_filename2


@lib.deprecated
def process(func, plugins, context, test=None):
    r"""Primary processing logic

    Takes callables and data as input, and performs
    logical operations on them until the currently
    registered test fails.

    If `plugins` is a callable, it is called early, before
    processing begins. If `context` is a callable, it will
    be called once per plug-in.

    Arguments:
        func (callable): Callable taking three arguments;
             plugin(Plugin), context(Context) and optional
             instance(Instance). Each must provide a matching
             interface to their corresponding objects.
        plugins (list, callable): Plug-ins to process. If a
            callable is provided, the return value is used
            as plug-ins. It is called with no arguments.
        context (Context, callable): Context whose instances
            are to be processed. If a callable is provided,
            the return value is used as context. It is called
            with no arguments.
        test (callable, optional): Provide custom test, defaults
            to the currently registered test.

    Yields:
        A result per complete process. If test fails,
        a TestFailed exception is returned, containing the
        variables used in the test. Finally, any exception
        thrown by `func` is yielded. Note that this is
        considered a bug in *your* code as you are the one
        supplying it.

    """

    __plugins = plugins
    __context = context

    if test is None:
        test = logic.registered_test()

    if hasattr(__plugins, "__call__"):
        plugins = __plugins()

    def gen(plugin, instances):
        if plugin.__instanceEnabled__ and len(instances) > 0:
            for instance in instances:
                yield instance
        else:
            yield None

    vars = {
        "nextOrder": None,
        "ordersWithError": list()
    }

    # Clear introspection values
    # TODO(marcus): Return *next* pair, this currently
    #   returns the current pair.
    self = process
    self.next_plugin = None
    self.next_instance = None

    for Plugin in plugins:
        self.next_plugin = Plugin
        vars["nextOrder"] = Plugin.order

        if not test(**vars):
            if hasattr(__context, "__call__"):
                context = __context()

            args = get_arg_spec(Plugin.process).args

            # Backwards compatibility with `asset`
            if "asset" in args:
                args.append("instance")

            instances = logic.instances_by_plugin(context, Plugin)

            # Limit processing to plug-ins with an available instance
            if not instances and "*" not in Plugin.families:
                continue

            for instance in gen(Plugin, instances):
                if instance is None and "instance" in args:
                    continue

                # Provide introspection
                self.next_instance = instance

                try:
                    result = func(Plugin, context, instance)

                except Exception as exc:
                    # Any exception occuring within the function
                    # you pass is yielded, you are expected to
                    # handle it.
                    yield exc

                else:
                    # Make note of the order at which
                    # the potential error error occured.
                    if result["error"]:
                        if Plugin.order not in vars["ordersWithError"]:
                            vars["ordersWithError"].append(Plugin.order)
                    yield result

            # Clear current
            self.next_instance = None

        else:
            yield logic.TestFailed(test(**vars), vars)
            break


process.next_plugin = None
process.next_instance = None

logic.process = process
